import asyncio
import logging
from bitstring import BitArray

from aiotorrent.core.response_handler import PeerResponseHandler as Handler
from aiotorrent.core.response_parser import PeerResponseParser as Parser
from aiotorrent.core.message_generator import MessageGenerator as Generator


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Peer:
	def __init__(self, address, torrent_info, priority = 10):
		self.address = address
		self.torrent_info = torrent_info

		self.active = False
		self.priority = priority
		# self.busy = False
		self.total_disconnects = 0

		self.choking_me = True
		self.am_interested = False
		self.has_handshaked = False
		self.has_bitfield = False

		# create empty BitArray of length equal to total number of pieces in the torrent
		num_pieces = len(torrent_info['piece_hashmap'])
		self.pieces = BitArray(num_pieces)


	def __repr__(self):
		return f"Peer({self.address})"
	

	def __lt__(self, other):
		return self.priority < other.priority


	async def connect(self):
		ip, port = self.address
		try:
			# creating connection variable for readability
			connection = asyncio.open_connection(ip, port)
			self.reader, self.writer = await asyncio.wait_for(connection, timeout=3)
			self.active = True
			logger.debug(f"Opened Connection to {self}")

		# ConnectionRefusedError: [WinError 1225] The remote computer refused the network connection
		# ConnectionResetError: [WinError 10054] An existing connection was forcibly closed by the remote host
		# ConnectionAbortedError: [WinError 10053] An established connection was aborted by the software in your host machine
		except(ConnectionRefusedError, ConnectionResetError, ConnectionAbortedError, OSError):
			await self.disconnect(f"Connection Refused/Reset/Aborted in CONNECT!")

		except asyncio.TimeoutError: await self.disconnect("Timed out while connecting!")


	async def disconnect(self, message=''):
		self.active = False
		self.total_disconnects += 1
		if hasattr(self, 'writer'):
			await self.writer.drain()
			self.writer.close()
			await self.writer.wait_closed()
		logger.debug(f"{self} {message} Closed Connnection")


	async def handshake(self):
		# send handshake after opening a connection successfully
		if self.active:
			ih = self.torrent_info['info_hash']
			handshake_message = Generator.gen_handshake(ih)
			response = await self.send_message(handshake_message)
			artifacts = Parser(response).parse()
			await Handler(artifacts, Peer=self).handle()



	async def intrested(self):
		# send intrested message if handshake is done and client is choked
		if self.active and self.has_handshaked:# and not self.choking_me:
			interested_message = Generator.gen_interested()
			response = await self.send_message(interested_message)
			artifacts = Parser(response).parse()
			await Handler(artifacts, Peer=self).handle()


	async def send_message(self, message, timeout=3):
		# Raise error if send_message() is called but peer is inactive
		# If send_message was called and the current status of the peer is not active
		# This means that this peer dropped the connection mid execution
		# So, we will re-establish a connection and re-raise the exception
		if not self.active:
			if self.total_disconnects > 10:
				return
			await self.connect()
			await self.handshake()
			await self.intrested()

			if self.active:
				logger.warning(f"Tried sending message to inactive {self}. Successfully re-established connection!")
			else:
				logger.warning(f"Tried sending message to inactive {self}. Failed to re-establish connection!")

			# Now raise BrokenPipeError so that the caller of send_message() can handle it
			raise BrokenPipeError(f"Tried sending message to inactive peer")

		if not self.active:
			raise BrokenPipeError(f"Connection to {self} has been closed")
			
		EMPTY_RESPONSE_THRESHOLD = 5
		response_buffer = bytes()
		self.writer.write(message)
		try:
			while True:
				response = await asyncio.wait_for(self.reader.read(1024), timeout=timeout)
				response_buffer += response

				logger.debug(f"{self}, {response=}")
				if len(response) <= 0: EMPTY_RESPONSE_THRESHOLD -= 1
				if EMPTY_RESPONSE_THRESHOLD < 0:
					await self.disconnect(f"Empty Response Threshold Exceeded!")

		# Timeout here is intentional and guaranteed. This is done to
		# receive full message because message gets sent in parts
		except asyncio.TimeoutError:
			...

		# ConnectionRefusedError: [WinError 1225] The remote computer refused the network connection
		# ConnectionAbortedError: [WinError 10053] An established connection was aborted by the software in your host machine
		# ConnectionResetError: [WinError 10054] An existing connection was forcibly closed by the remote host
		except(ConnectionRefusedError, ConnectionResetError, ConnectionAbortedError):
			await self.disconnect(f"Connection Refused/Reset/Aborted in SEND!")

		finally:
			return response_buffer


	def update_piece_info(self, piece_num: int, has_piece: bool):
		# Utility function to update piece information of peer
		self.pieces[piece_num] = has_piece






import os
import asyncio
import logging
from struct import unpack
from ipaddress import IPv4Address

# import bencode
from aiotorrent.core.bencode_utils import bencode_util
# from bencode._bencode import BTFailure

from aiotorrent.core.util import chunk


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class DHTProtocolHelper(asyncio.DatagramProtocol):
    def __init__(self, message, on_response, on_error=None):
        self.message = message
        self.on_response = on_response
        self.on_error = on_error or (lambda e: None)

    def connection_made(self, transport):
        self.transport = transport
        self.transport.sendto(self.message)

    def datagram_received(self, data, addr):
        self.on_response(data)
        self.transport.close()

    def error_received(self, exc):
        self.on_error(exc)
        self.transport.close()

    def connection_lost(self, exc):
        pass  # Optional: handle cleanup


class SimpleDHTCrawler:
    FOUND_PEERS = set()
    _nodes_to_crawl = asyncio.LifoQueue()

    def __init__(self, info_hash, node_id = None, bootstrap_nodes = []):
        self.node_id = node_id or os.urandom(20)
        self.info_hash = info_hash

        self.bootstrap_nodes = bootstrap_nodes or  [
            ('router.bittorrent.com', 6881),
            ('router.utorrent.com', 6881),
            ('trdht.transmissionbt.com', 6881),
        ]

        for node in self.bootstrap_nodes:
            self._nodes_to_crawl.put_nowait(node)


    def _generate_get_peers_query(self, transaction_id, info_hash):
        query = {
            b't': transaction_id,
            b'y': b'q',
            b'q': b'get_peers',
            b'a': {
                b'id': self.node_id,
                b'info_hash': info_hash
            }
        }
        return bencode_util.bencode(query)


    def _bytes_to_address(self, blob):
        if isinstance(blob, str):
            blob = blob.encode()
        try:
            ip, port = unpack('>IH', blob)
            ip = IPv4Address(ip).compressed
            return (ip, port)
        # except BTFailure as e:
        #     logger.error(f"Invalid IP Address {ip:port}: {e}")
        except Exception as e:
            logger.error(f"An unknown error occured decoding IP Address {ip:port}: {e}")


    def _decode_nodes(self, nodes_blob):
        nodes = []
        for node_info in chunk(nodes_blob, 26):
            node_id = node_info[:19]
            node_ip, node_port = self._bytes_to_address(node_info[20:26])
            # I mean do we really need the node_id?
            nodes.append((node_ip, node_port))
        
        return nodes


    def parse_response(self, response):
        peers = []
        closer_nodes = []
        try:
            response = bencode_util.bdecode(response)
            if not response or 'r' not in response:
                return None

            # Found peers directly
            if 'values' in response['r']:
                for peer_addr in response['r']['values']:
                    peers.append(self._bytes_to_address(peer_addr))
            
            if 'nodes' in response['r']:
                closer_nodes_blob = response['r']['nodes']
                closer_nodes.extend(self._decode_nodes(closer_nodes_blob))

        # except BTFailure as e:
        #     logger.error(f"Error decoding bencoded data recieved from peer {peer_addr}: {e}")
        
        except Exception as e:
            logger.error(f"An unknown error occured while parsind bencoded data recieved from {peer_addr}: {e}")

        finally:
            return (peers, closer_nodes)


    async def send_get_peers_req(self, peer_addr, message, loop, _semaphore, timeout=5):
        response_future = loop.create_future()

        def on_response(data):
            peers, closer_nodes = self.parse_response(data)
            if peers:
                self.FOUND_PEERS  |= set(peers)

            for node in closer_nodes:
                self._nodes_to_crawl.put_nowait(node)

            if not response_future.done():
                response_future.set_result(data)

        def on_error(exc):
            if not response_future.done():
                response_future.set_exception(exc)

        try:
            await loop.create_datagram_endpoint(
                lambda: DHTProtocolHelper(message, on_response, on_error),
                remote_addr=peer_addr
            )
            return await asyncio.wait_for(response_future, timeout)

        except asyncio.TimeoutError:
            logger.debug(f"Timeout from {peer_addr}")
            return None
        
        except Exception as e:
            logger.error(f"An unknown error occured while sending a datagram to {peer_addr}: {e}")
        
        finally:
            _semaphore.release()


    async def crawl(self, min_peers_to_retrieve = 100, max_connections = 256):
        loop = asyncio.get_running_loop()
        logger.info(f"Starting DHT crawl with Node ID: {self.node_id.hex()}")

        processed_count = 0
        semaphore = asyncio.Semaphore(max_connections)

        while len(self.FOUND_PEERS) < min_peers_to_retrieve:
            if self._nodes_to_crawl.empty():
                logger.info("Exhausted all available nodes on DHT")
                break
            
            await semaphore.acquire()
            peer_addr = await self._nodes_to_crawl.get()
            transaction_id = os.urandom(2)
            message = self._generate_get_peers_query(transaction_id, self.info_hash)
            asyncio.create_task(self.send_get_peers_req(peer_addr, message, loop, semaphore))
            processed_count += 1
            await asyncio.sleep(0.5)

            # Empty queue to check how the program handles an exhausted queue
            # if len(self.FOUND_PEERS) > min_peers_to_retrieve / 2:
            #     while not self._nodes_to_crawl.empty():
            #         self._nodes_to_crawl.get_nowait()

            logger.info(f"Found {len(self.FOUND_PEERS)}/{min_peers_to_retrieve} peers [{self._nodes_to_crawl.qsize()} in queue]")

        logger.info(f"Found {len(self.FOUND_PEERS)} peers after crawling {processed_count} nodes")
        logger.debug(f"{self._nodes_to_crawl.qsize()} nodes left in queue")
        
        return self.FOUND_PEERS
        

if __name__ == "__main__":
    INFO_HASH = b""
    node_id = os.urandom(20)

    dht_crawler = SimpleDHTCrawler(INFO_HASH, node_id)
    asyncio.run(dht_crawler.crawl())


from struct import pack


class MessageGenerator:
	"""
		This class generates messages.
	"""
	@staticmethod
	def gen_handshake(info_hash):
		'''
		handshake:
			<pstrlen>	1   byte; string length of <pstr>, as a single raw byte; strlen = 19
			<pstr>		19 bytes; string identifier of the protocol; "BitTorrent protocol"
			<reserved>	8  bytes; eight (8) reserved bytes
			<info_hash>	20 bytes; info_hash
			<peer_id>	20 bytes; peer_id
		'''
		message = pack(
			">B19sQ20s20s",
			19,
			b"BitTorrent protocol",
			00000000,
			info_hash,
			b"ABCD" + b"X"*16			
		)

		return message


	@staticmethod
	def gen_interested():
		mlen, mid = 1, 2
		message = pack(">IB", mlen, mid)
		return message


	@staticmethod
	def gen_request(index, offset, BLOCK_SIZE=(2 ** 14)):
		mlen, mid = 13, 6
		message = pack(">IBIII", mlen, mid, index, offset, BLOCK_SIZE)
		return message


import asyncio
import argparse
import logging

from aiotorrent.aiotorrent import Torrent

"""
    aiotorrent download <path/to/file.torrent> <save_location<optional>>
    aiotorrent stream <path/to/file.torrent>
"""

LOG_LEVELS = {
    0: logging.NOTSET,
    1: logging.CRITICAL,
    2: logging.ERROR,
    3: logging.WARN,
    4: logging.INFO,
    5: logging.DEBUG,
}


async def download_torrent(torrent_file_loc, save_loc=None):
    # TODO: Add parameter for save location
    # TODO: Add parameter for download strategy
    torrent = Torrent(torrent_file_loc)
    await torrent.init(dht_enabled = True)
    for file in torrent.files:
        await torrent.download(file)


async def stream_torrent(torrent_file_loc, host="127.0.0.0", port=8080):
    torrent = Torrent(torrent_file_loc)
    await torrent.init(dht_enabled = True)
    for file in torrent.files:
        await torrent.stream(file, host=host, port=port)


def print_torrent_info(torrent_file_loc, format="json", verbose=False):
    torrent = Torrent(torrent_file_loc)
    info = torrent.get_torrent_info(format=format, verbose=verbose)
    print(info)



# ================================ Create parsers here ================================
def create_download_parser(subparsers):
    download_parser = subparsers.add_parser(
        "download",
        help="Download torrent files"
    )
    download_parser.add_argument(
        "torrent_path",
        help="Path to the .torrent file"
    )
    download_parser.add_argument(
        "save_location",
        nargs="?", # This makes the argument optional (0 or 1 argument)
        default=".", 
        help="(Optional) Save downloaded files to location (defaults to current directory)"
    )
    
    return download_parser

def create_stream_parser(subparsers):
    stream_parser = subparsers.add_parser(
        "stream",
        help="Stream files over HTTP"
    )
    stream_parser.add_argument(
        "torrent_path",
        help="Path to the .torrent file"
    )
    # Add opional arguments for host and port
    stream_parser.add_argument(
        "--host",
        "-H", # Short form
        default="localhost", # Default value if not provided
        help="Host address for streaming (defaults to localhost)"
    )
    stream_parser.add_argument(
        "--port",
        "-p", # Short form
        type=int,
        default=8080, # Default value if not provided
        help="Port for streaming (defaults to 8080)"
    )

    return stream_parser


def create_info_parser(subparsers):
    info_parser = subparsers.add_parser(
        "info",
        help="Parse and show torrent metadata"
    )
    info_parser.add_argument(
        "torrent_path",
        help="Path to the .torrent file"
    )
    info_parser.add_argument(
        "--format",
        "-f",
        default="json",
        help="Format to print the torrent metadata. Defaults to json"
    )

    return info_parser


# ================================ Main parser ================================
async def main_parser():
    """Sets up and parses the arguments, then calls the appropriate function."""
    parser = argparse.ArgumentParser(
        description="aiotorrent CLI for downloading and streaming torrents."
    )
    # Add the verbosity argument here
    parser.add_argument(
        '-v', '--verbose',
        action='count',
        default=0,  # Default verbosity is 0 if no -v is given
        help='Show verbose output. Use -vv, -vvv, etc for increased verbosity'
    )

    # Create subparsers for the different commands
    subparsers = parser.add_subparsers(
        dest="command",
        help="Available commands"
    )

    download_parser = create_download_parser(subparsers)
    stream_parser = create_stream_parser(subparsers)
    info_parser = create_info_parser(subparsers)

    download_parser.set_defaults(func=download_torrent)
    stream_parser.set_defaults(func=stream_torrent)
    info_parser.set_defaults(func=print_torrent_info)
    args = parser.parse_args()

    # Check if verbosity flag is set, if not it is set to 0
    # If set, ensure the verbosity doesn't exceed level 5 (DEBUG)
    verbosity = min(5, args.verbose) if args.verbose else 0
    log_level = LOG_LEVELS[verbosity]

    logging.basicConfig(level=log_level, handlers=[
        logging.StreamHandler(),
    ])


    if hasattr(args, 'func'):
        if args.command == 'download':
            await args.func(args.torrent_path, args.save_location)
        elif args.command == 'stream':
            await args.func(args.torrent_path, args.host, args.port)
        elif args.command == 'info':
            args.func(args.torrent_path, args.format, args.verbose)
    else:
        parser.print_help()


def main():
    asyncio.run(main_parser())


if __name__ == "__main__":
    main()


import io
import copy
import asyncio
import hashlib
import logging
import platform

import json
from aiotorrent.peer import Peer
from aiotorrent.core.bencode_utils import bencode_util
from aiotorrent.core.util import chunk, PieceWriter
from aiotorrent.core.file_utils import FileTree
from aiotorrent.tracker_factory import TrackerFactory
from aiotorrent.downloader import FilesDownloadManager
from aiotorrent.core.util import DownloadStrategy
from aiotorrent.DHTv4 import SimpleDHTCrawler


# Asyncio throws runtime error if the platform is windows
# RuntimeError: Event loop is closed
# FOUND FIX: https://stackoverflow.com/a/66772242
if platform.system() == 'Windows':
	asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Torrent:
	def __init__(self, torrent_file):
		# The object passed is a file-like object
		if isinstance(torrent_file, io.IOBase):
			bencoded_data = torrent_file.read()
		else:
			# The object passed is a filepath
			with open(torrent_file, 'rb') as torrent:
				bencoded_data = torrent.read()

		# dict_keys(['files', 'name', 'piece length', 'pieces'])
		data = bencode_util.bdecode(bencoded_data)

		self.trackers = list()
		self.peers = list()
		self.name = data['info']['name']
		self.files = None # This will be replaced with a file_tree object

		# Check if this torrent has multiple files
		self.has_multiple_files = True if 'files' in data['info'] else False

		size = int()
		peers = list()
		trackers = list()
		piece_hashmap = dict()
		announce = data['announce'] if 'announce' in data else None # announce is a string
		files = data['info']['files'] if self.has_multiple_files else self.name
		piece_len = data['info']['piece length']

		# If torrent has multiple files, set torrent
		# size to sum of length of all individual files
		if self.has_multiple_files:
			size = sum([file['length'] for file in files])
		else:
			size = data['info']['length']

		# re-encode info to -> bencode and then apply sha-1 to it
		raw_info_hash = bencode_util.bencode(data['info'])
		info_hash = hashlib.sha1(raw_info_hash).digest()

		# get pieces and convert it from hexadecimal (string) to bytes object
		raw_pieces = data['info']['pieces']

		# for every 20 byte in raw_pieces -> str, map index:piece (int:str) to pieces
		for index, piece in enumerate(chunk(raw_pieces, 20)):
			piece_hashmap[index] = piece

		self.torrent_info = {
			'name': data['info']['name'],
			'size': size,
			'files': files,
			'piece_len': piece_len,
			'info_hash': info_hash,
			'piece_hashmap': piece_hashmap,
			'peers': peers,
			'trackers': trackers,
		}


		# Add announce url to trackers list if announce exists
		if announce: self.torrent_info['trackers'].append(announce)

		# Adding trackers to torrent info. Tracker is a list that contains a string
		if 'announce-list' in data:
			for tracker in data['announce-list']:
				tracker = tracker[0]
				if not tracker in self.torrent_info['trackers']:
					self.torrent_info['trackers'].append(tracker)

		self.files = FileTree(self.torrent_info)

		# for file in self.files:
		# 	logger.debug(f"File: {file}")

			
	async def _contact_trackers(self):
		task_list = list()

		for tracker_addr in self.torrent_info['trackers']:
			tracker = TrackerFactory(tracker_addr, self.torrent_info)
			self.trackers.append(tracker)
			task_list.append(asyncio.create_task(tracker.get_peers()))

		await asyncio.gather(*task_list)


	def _get_peers(self):
		peers_aggregated = set()
		# Get peers address from each tracker
		# for tracker in self.trackers:
		# 	for peer in tracker.peers:
		# 		if not peer in self.torrent_info['peers']:
		# 			# Add peer address to torrent info
		# 			self.torrent_info['peers'].append(peer)
		# 			# create and add peers object to self
		# 			self.peers.append(Peer(peer, self.torrent_info))

		for tracker in self.trackers:
			peer_list = set(tracker.peers)
			peers_aggregated |= peer_list

		logger.info(f"Got {len(peers_aggregated)} Peers for this torrent")
		return peers_aggregated


	async def _get_peers_dht(self, timeout = 30):
		info_hash = self.torrent_info['info_hash']
		dht_crawler = SimpleDHTCrawler(info_hash)
		peers = await dht_crawler.crawl(min_peers_to_retrieve=100)
		# peers = await asyncio.wait_for(dht_crawler.crawl(), timeout = timeout)
		logger.info(f"Got {len(peers)} Peers using DHT")
		return peers


	def show_files(self):
		for file in self.files:
			logger.info(f"File: {file}")


	async def init(self, dht_enabled = False):
		# Contact Trackers and get peers
		await self._contact_trackers()
		peer_addrs = self._get_peers() #TODO: Rename get_peers to add_peers

		dht_peers = set()
		if dht_enabled:
			dht_peers = await self._get_peers_dht(timeout=30)
			peer_addrs |= dht_peers

		# If there are more than 512 peers, randomly shortlist them
		#TODO: Use a better strategy to shortlist peers. 
		# When I tried opening connections to ~1300 peers parallely
		# Error: ValueError: too many file descriptors in select() 
		# ConnectionResetError: [WinError 10054] An existing connection was forcibly closed by the remote host
		# if len(peer_addrs) > 128:
		# 	shortlisted_peers = set()
		# 	for _ in range(128):
		# 		shortlisted_peers.add(peer_addrs.pop())

		# Initialize and attach all aggregated peers objects to self
		self.peers = [Peer(peer, self.torrent_info) for peer in peer_addrs]

		# Use list comprehension to create and execute peer functions in parallel
		connections = [peer.connect() for peer in self.peers]
		await asyncio.gather(*connections)

		handshakes = [peer.handshake() for peer in self.peers]
		await asyncio.gather(*handshakes)

		interested_msgs = [peer.intrested() for peer in self.peers]
		await asyncio.gather(*interested_msgs)

		# Add peers addresses to torrent_info
		self.torrent_info['peers'] = peer_addrs

		# Info
		active_peers = [peer for peer in self.peers if peer.has_handshaked]
		active_trackers = [tracker for tracker in self.trackers if tracker.active]
		logger.info(f"{len(active_peers)} peers active")
		logger.info(f"{len(active_trackers)} trackers active")


	async def download(self, file, strategy=DownloadStrategy.DEFAULT):
		#TODO: Add a peer_list parameter with the default value of self.peers
		active_peers = [peer for peer in self.peers if peer.has_handshaked]
		fd_man = FilesDownloadManager(self.torrent_info, active_peers)
		directory = self.torrent_info['name']
		logger.info(f"Using strategy {strategy} to download file {file}")

		with PieceWriter(directory, file) as piece_writer:
			if strategy == DownloadStrategy.DEFAULT:
				async for piece in fd_man.get_file(file):
					piece_writer.write(piece)

			elif strategy == DownloadStrategy.SEQUENTIAL:
				piece_len = self.torrent_info['piece_len']
				async for piece in fd_man.get_file_sequential(file, piece_len):
					piece_writer.write(piece)


	async def __generate_torrent_stream(self, file):
		active_peers = [peer for peer in self.peers if peer.has_handshaked]
		fd_man = FilesDownloadManager(self.torrent_info, active_peers)
		piece_len = self.torrent_info['piece_len']
		async for piece in fd_man.get_file_sequential(file, piece_len):
			yield piece.data


	async def stream(self, file, host="127.0.0.1", port=8080):
		try:
			from starlette.applications import Starlette
			from starlette.responses import StreamingResponse
			from starlette.routing import Route
			import uvicorn
		except (ImportError, ModuleNotFoundError):
			raise ModuleNotFoundError("Streaming dependencies not found")

		async def homepage(request):
			return StreamingResponse(
				self.__generate_torrent_stream(file),
				media_type='video/mp4'
			)

		app = Starlette(debug=True, routes=[
				Route('/', homepage),
		])

		config = uvicorn.Config(app, host=host, port=port)
		server = uvicorn.Server(config)
		await server.serve()


	def get_torrent_info(self, format='json', verbose=False):
		# TODO: Add Yaml as an export format
		torrent_info = copy.deepcopy(self.torrent_info)
		torrent_info['info_hash'] = torrent_info['info_hash'].hex()
		
		piece_hashmap = torrent_info.pop('piece_hashmap')
		peer_list = torrent_info.pop('peers')

		if verbose:
			# Serialize piece hashes by converting byte objects to their hex representations
			torrent_info['piece_hashmap'] = {}
			for piece_no, piece_hash in piece_hashmap.items():
				torrent_info['piece_hashmap'][piece_no] = piece_hash.hex()

			torrent_info['peers'] = tuple(peer_list)

		return json.dumps(torrent_info)



class File:
	#TODO: Rename to TorrentFile
	"""
	{
		'length': int,
		'path': list[str]
	}
	"""
	def __init__(self, file_info: dict, counted: int, piece_size: int) -> None:
		self.size = file_info['length']

		# In case of single file torrent, the name is a string, and not a list of filenames[strings]
		filename = file_info['path']
		self.name = filename[0] if isinstance(filename, list) else filename
		self.__bytes_written = 0
		self.__bytes_downloaded = 0

		start_piece, start_byte = divmod(counted, piece_size)
		end_piece, end_byte = divmod((counted + self.size), piece_size)

		self.start_piece = start_piece
		self.start_byte = start_byte
		self.end_piece = end_piece
		self.end_byte = end_byte


	def __repr__(self):
		return f"{self.name} ({self.size})"
	

	def get_bytes_written(self):
		"""Returns the number of bytes written to disk for the file"""
		return self.__bytes_written


	def _set_bytes_written(self, value):
		self.__bytes_written = value


	def get_bytes_downloaded(self):
		"""
			Returns the number of bytes downloaded for the file, regardless of whether they were written to disk or not.
			Bytes (Pieces) may be discarded if the piece is declared invalid (Piece hash does not match).
		"""
		return self.__bytes_downloaded
	

	def _set_bytes_downloaded(self, value):
		self.__bytes_downloaded = value


	def get_download_progress(self, precision=2):
		"""Convenience function for getting the download progress in percentage"""
		download_progress = (self.__bytes_written / self.size) * 100
		return round(download_progress, precision)


class FileTree(list):
	"""
	FileTree inherits from list class making FileTree objects
	an iterator as well as making them subscriptable.
	"""
	def __init__(self, torrent_info: dict) -> None:
		counted = 0
		piece_size = torrent_info['piece_len']

		# If it is a single file torrent, create a dictionary in a torrent format
		if not isinstance(torrent_info['files'], list):
			file = {
				'path': torrent_info['files'],
				'length': torrent_info['size'],
			}
			self.append(File(file, counted, piece_size))
			return

		# Create file object for every file in the torrent.
		# Increase the value of the counted var by length of each file.
		for file in torrent_info['files']:
			self.append(File(file, counted, piece_size))
			counted += file['length']




import asyncio
import logging
from pathlib import Path

from aiotorrent.piece import Piece
from aiotorrent.core.util import BLOCK_SIZE
from aiotorrent.core.file_utils import File, FileTree
from aiotorrent.core.util import SequentialPieceDispatcher

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class FilesDownloadManager:
	def __init__(self, torrent_info: dict, active_peers: list):
		# Extract torrent size and piece size values from torrent info
		piece_size = torrent_info['piece_len']
		torrent_size = torrent_info['size']

		# Create a directory with the same name as torrent name to download files to
		self.directory = torrent_info['name']
		Path(self.directory).mkdir(exist_ok=True)

		# divmod returns a tuple which is the quotient and remainder
		total_pieces, last_piece = divmod(torrent_size, piece_size)
		total_blocks, last_block = divmod(piece_size, BLOCK_SIZE)
		
		# Increment total number of pieces if last piece exists
		if last_piece: total_pieces += 1
		if last_block: total_blocks += 1

		# Since we're taking into account zero based piece index
		total_pieces -= 1

		piece_info = {
			'total_pieces': total_pieces,
			'total_blocks': total_blocks,
			'last_piece': last_piece,
			'last_block': last_block
		}

		self.piece_info = piece_info
		self.piece_hashmap = torrent_info['piece_hashmap']
		self.file_tree = FileTree(torrent_info)

		peer_def = 10   # Peer default priority
		self.peer_queue = asyncio.PriorityQueue()
		for peer in active_peers:
			self.peer_queue.put_nowait((peer_def, peer))

		# Queue for storing pieces for a file
		self.file_pieces = asyncio.PriorityQueue()
		


	def create_pieces_queue(self, file: File) -> None:
		piece_def = 3   # Default piece priority
		for piece_num in range(file.start_piece, file.end_piece + 1):
			self.file_pieces.put_nowait((piece_def, piece_num))


	def file_downloaded(self) -> bool:
		'''
		Returns true if all the pieces have been downloaded, false otherwise
		'''
		return True if self.file_pieces.empty() else False


	async def get_file(self, file: File) -> Piece:
		self.create_pieces_queue(file)
		task_list = list()

		while not self.file_downloaded():
			prio_piece, num = await self.file_pieces.get()
			piece = Piece(num, prio_piece, self.piece_info)
			task = asyncio.create_task(piece.download(self.peer_queue))
			task_list.append(task)

		for task in asyncio.as_completed(task_list):
			piece = await task

			if not Piece.is_valid(piece, self.piece_hashmap):
				self.file_pieces.put_nowait((1, piece.num))
				continue

			if file.start_piece == piece.num:
				piece.data = piece.data[file.start_byte:]

			if file.end_piece == piece.num:
				piece.data = piece.data[:file.end_byte]

			file._set_bytes_written(file.get_bytes_written() + len(piece.data))
			yield piece

		logger.info(f"File {file} downloaded")


	async def get_file_sequential(self, file: File, piece_len) -> Piece:
		task_list = []
		dispatch_manager = SequentialPieceDispatcher(file, piece_len)

		self.create_pieces_queue(file)
		max_concurrent_pieces = 10
		sema = asyncio.Semaphore(max_concurrent_pieces)

		while not self.file_downloaded():
			async with sema:
				await sema.acquire()
				prio_piece, num = await self.file_pieces.get()
				piece = Piece(num, prio_piece, self.piece_info)
				# Passing semaphore so that piece can release it when it finishes fetching
				# all the blocks for itself
				task = asyncio.create_task(piece.download(self.peer_queue, _semaphore=sema))
				task_list.append(task)

				for task in task_list:
					if not task.done():
						break
					else:
						task_list.remove(task)
						piece = task.result()
						file._set_bytes_downloaded(file.get_bytes_downloaded() + len(piece.data))
						# yield piece
	
						if not Piece.is_valid(piece, self.piece_hashmap):
							self.file_pieces.put_nowait((1, piece.num))
							continue

						if file.start_piece == piece.num:
							piece.data = piece.data[file.start_byte:]

						if file.end_piece == piece.num:
							piece.data = piece.data[:file.end_byte]
							
						await dispatch_manager.put(piece)
						async for piece in dispatch_manager.dispatch():
							file._set_bytes_written(file.get_bytes_written() + len(piece.data))
							yield piece

		async for piece in dispatch_manager.drain():
			file._set_bytes_written(file.get_bytes_written() + len(piece.data))
			yield piece



import logging
from struct import unpack
from struct import error as UnpackError


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PeerResponseParser:
	def __init__(self, response):
		self.response = response
		
		self.messages = {
		# self created index. Does not necessarily follow bittorrent protocol specification
			0: self.parse_choke,
			1: self.parse_unchoke,
			4: self.parse_have,
			5: self.parse_bitfield,
			7: self.parse_piece,
			19: self.parse_handshake,
			None: self.parse_keep_alive
		}
		self.artifacts = dict()


	def parse(self):

		while self.response:
			try:
				# if first byte of response is 19 bytes, it's a handshake
				if self.response[0] == 19:
					self.parse_handshake()
					continue #	issue #1

				# message len and id need to be class wide accessible because message len
				# needs to be updated in case of a bitfield and piece requests
				self.message_len = unpack('>I', self.response[:4])[0]
				self.message_id = unpack('>B', self.response[4:5])[0] if self.message_len != 0 else None
					
				# keep-alive messages have length 0 and no message_id
				if self.message_len == 0 and not self.message_id: self.parse_keep_alive()

				# if msg id not in message index, clear response
				if self.message_id not in self.messages:
					logger.warning(f"{self.message_id=}, {self.message_len=}, {self.response[:16]}")
					self.response = bytes()

				# finally parse the blob of response
				logger.debug(f"{self.message_len=}, {self.message_id=}, {self.response[:16]=}")
				self.messages[self.message_id]()

			# In case of general exception, clear the response
			except Exception as E:
				logger.warning(f"Parser: {E}")
				self.response = bytes()
			
		return self.artifacts

		
	def parse_keep_alive(self):
		# keep-alive
		message = self.response[:4]
		self.response = self.response[4:]
		self.artifacts.update({'keep_alive': True})
		
	
	def parse_choke(self):
		# client got choked by peer
		message = self.response[:5]
		self.response = self.response[5:]
		self.artifacts.update({'choke': True})
		
		
	def parse_unchoke(self):
		message = self.response[:5]
		self.response = self.response[5:]
		self.artifacts.update({'unchoke': True})
		
		
	def parse_have(self):
		message = self.response[:9]
		# We just need the piece index so we can ignore the first 5 bytes of message.
		# Unpack returns tuple , selecting first element.
		piece_index = unpack('>I', message[5:])[0]
		self.response = self.response[9:]
		self.artifacts.update({'have': {piece_index: True}})
		
	
	def parse_piece(self):
		# Returns index, offset and block except when there's
		# an unpack error. In the second case, It raises a TypeError
		if not 'pieces' in self.artifacts: self.artifacts['pieces'] = list()
		block_len = self.message_len - 9
		total = block_len + 13
		try:
			index, offset = unpack('>II', self.response[5: 13])
			data = self.response[13:total]
			block_info = (index, offset, data)
			self.artifacts['pieces'].append(block_info)
		except UnpackError:
			raise TypeError("Parser: Failed to extract piece")
		finally:
			self.response = self.response[total:]
		
		
	def parse_bitfield(self):
		# Since bitfield len is 1+X. X is length of bitfield
		self.message_len -= 1
		total = self.message_len + self.message_id
		message = self.response[5:total]
		self.response = self.response[total:]
		self.artifacts.update({'bitfield': message})
		
		
	def parse_handshake(self):
		# Total is 68 bytes
		total = 68
		message = self.response[:total]
		self.response = self.response[total:]
		self.artifacts.update({'handshake': message})

		

if __name__ == "__main__":
	...


import asyncio
import hashlib
import logging

from aiotorrent.core.util import Block
from aiotorrent.core.response_parser import PeerResponseParser as Parser
from aiotorrent.core.response_handler import PeerResponseHandler as Handler
from aiotorrent.core.message_generator import MessageGenerator as Generator

from aiotorrent.core.util import BLOCK_SIZE, BLOCKS_PER_CYCLE, MIN_BLOCKS_PER_CYCLE


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

#TODO: Update class paramater documentation
#TODO: Rename peers_man to peer_queue
#TODO: Use consistent naming convention for other variables
class Piece:
	def __init__(self, num: int, priority: int, piece_info: dict[str, int]):
		"""
		num: int
			Zero based piece index to be fetched from peers

		piece_info: dict
			dictionary containing information regarding pieces

		peers_man: PeersManager
			Object of PeersManager
		"""
		self.data = bytes()
		self.blocks = dict()
		self.num = num
		self.priority = priority

		self._is_last_piece = False
		self.total_blocks = piece_info['total_blocks']
		self.piece_size = self.total_blocks * BLOCK_SIZE

		# If num matches the num of total pieces then it's the last piece.
		if self.num == piece_info['total_pieces']:
			self._is_last_piece = True
			self.total_blocks, self._last_offset = divmod(piece_info['last_piece'], BLOCK_SIZE)
			self.piece_size = (self.total_blocks * BLOCK_SIZE) + self._last_offset


	def __repr__(self):
		return (f"Piece #{self.num}")


	async def fetch_blocks(self, block_offsets: list[int], peer) -> list[Block]:
		"""
		This function fetches a block from a peer. It returns
		either an Block object or a None Object.
	
		block_offsets: list[int]
			Zero based block offsets which should be a multiple of
			BLOCK_SIZE. If there are 10 blocks in piece #0, block
			offset for block num 8 would be (8 * BLOCK_SIZE) = 131702
		"""
		requests = bytes()

		for offset in block_offsets:
			block_num = int(offset / BLOCK_SIZE)
			logger.debug(f"Requesting Block #{self.num}-{block_num} from {peer}")
			request_message = Generator.gen_request(self.num, offset)

			# Last block of last piece will be requested with second last block.
			is_last_block = True if block_num == (self.total_blocks - 1) else False

			if self._is_last_piece and is_last_block:
				request_message = Generator.gen_request(self.num, offset, BLOCK_SIZE + self._last_offset)

			requests += request_message

		response = await peer.send_message(requests, timeout=5)

		# If peer sends empty block, update the piece_info of peer
		# by setting it to false and raise IOError
		if not response:
			peer.update_piece_info(self.num, False)
			raise IOError(f"{peer} Sent Empty Blocks")

		try:
			artifacts = Parser(response).parse()
			blocks = await Handler(artifacts, Peer=peer).handle()
			for block in blocks:
				logger.debug(f"Got {block} from {peer}")

			return blocks
		
		except TypeError as E:
			logging.info(f"Requesting Blocks for {self} from {peer} Returned None")
			logging.warning(E)
			self.adjust_blocks_per_cycle(-1)
			return None


	def is_piece_complete(self) -> bool:
		for block_num in range(self.total_blocks):
			if not block_num in self.blocks:
				return False
		return True


	def gen_offsets(self) -> set:
		blocks = set()
		total_blocks = self.total_blocks
		if self._is_last_piece: total_blocks += 1
		for block_num in range(self.total_blocks):
			if not block_num in self.blocks:
				block_offset = block_num * BLOCK_SIZE
				blocks.add(block_offset)
		return blocks


	@staticmethod
	def is_valid(piece, piece_hashmap):
		#TODO: make this an instance method so that every piece can validate it's own data
		piece_hash = hashlib.sha1(piece.data).digest()

		if piece_hash != piece_hashmap[piece.num]:
			logging.warning(f"Piece Hash Does Not Match for {piece}")
			return False
			
		return True


	def adjust_blocks_per_cycle(self, value: int = 1):
		"""
		Increments / Decrements BLOCKS_PER_CYCLE by `value`
		"""
		global BLOCKS_PER_CYCLE
		BLOCKS_PER_CYCLE += value

		# Ensure BLOCKS_PER_CYCLE is always greater than MIN_BLOCKS_PER_CYCLE
		BLOCKS_PER_CYCLE = max(BLOCKS_PER_CYCLE, MIN_BLOCKS_PER_CYCLE)

		# Ensure BLOCKS_PER_CYCLE is always less than total_blocks in a piece
		BLOCKS_PER_CYCLE = min(BLOCKS_PER_CYCLE, self.total_blocks)


	async def download(self, peers_man, _semaphore = None) -> 'Piece':
		priority, peer = await peers_man.get()

		while not self.is_piece_complete():
			task_list = list()
			block_offsets = self.gen_offsets()

			# If number of block offsets to be downloaded is less than BLOCKS_PER_CYCLE,
			# all blocks can be downloaded in a single cycle.
			if len(block_offsets) >= BLOCKS_PER_CYCLE:
				offsets = {block_offsets.pop() for _ in range(BLOCKS_PER_CYCLE)}
				block_offsets.difference_update(offsets)
			else:
				offsets = self.gen_offsets()

			blocks = self.fetch_blocks(offsets, peer)
			task = asyncio.create_task(blocks)
			task_list.append(task)

			try:
				# Get results and increment BLOCKS_PER_CYCLE if successful
				results = await asyncio.gather(*task_list)
				self.adjust_blocks_per_cycle(1)

			except (BrokenPipeError, IOError):
				# If BrokenPipe or IOError recieved then we need to reduce the peers priority
				current_priority, current_peer = priority, peer
				priority, peer = await peers_man.get()
				await peers_man.put((current_priority + 1, current_peer))
				continue

			# Remove NoneType objects and merge inner lists to outerlists
			results = [result for result in results if result]
			results = sum(results, [])

			# In case of successful retrieval of block, add block to self.blocks
			for block in results:
				if block.data:
					self.blocks.update({block.num: block})

		# Concatenate all the block values
		for block_num in range(self.total_blocks):
			self.data += self.blocks[block_num].data

		await peers_man.put((priority - 1, peer))
		# Release semaphore so that the next task can begin
		if _semaphore is not None:
			_semaphore.release()
		return self



import fastbencode


class BencodeUtil:
    def to_py(self, x):
        if isinstance(x, dict):
            result = {}
            for k, v in x.items():
                new_k = k.decode('utf-8') if isinstance(k, bytes) else k
                result[new_k] = self.to_py(v)
            return result
        
        if isinstance(x, list):
            return [self.to_py(v) for v in x]
        
        if isinstance(x, bytes):
            try:
                return x.decode('utf-8')
            except UnicodeDecodeError:
                return x
        return x


    def to_bytes(self, x):
        if isinstance(x, dict):
            result = {}
            for k, v in x.items():
                new_k = k.encode('utf-8') if isinstance(k, str) else k
                result[new_k] = self.to_bytes(v)
            return result
        
        if isinstance(x, list):
            return [self.to_bytes(v) for v in x]

        if isinstance(x, str):
            return x.encode('utf-8')
        return x


    def bdecode(self, data):
        return self.to_py(fastbencode.bdecode(data))


    def bencode(self, obj):
        return fastbencode.bencode(self.to_bytes(obj))


bencode_util = BencodeUtil()


import logging
from struct import unpack
from bitstring import BitArray

from aiotorrent.core.util import Block


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PeerResponseHandler:
	def __init__(self, artifacts, Peer=None):
		self.artifacts = artifacts
		self.Peer = Peer


	async def handle(self):

		if logger.isEnabledFor(logging.DEBUG):
			for key, value in self.artifacts.items():
				if isinstance(value, bytes):
					logger.debug(f"{key}: {value[:32]}")

		while self.artifacts:
			if "keep_alive" in self.artifacts: self.handle_keep_alive() 
			if "choke" in self.artifacts: await self.handle_choke()
			if "unchoke" in self.artifacts: self.handle_unchoke()
			if "handshake" in self.artifacts: await self.handle_handshake()
			# Merged have_handler into bitfield_handler
			if "have" in self.artifacts: self.handle_bitfield()
			if "bitfield" in self.artifacts: self.handle_bitfield()
			# Piece handler is special as it returns values
			if "pieces" in self.artifacts: return self.handle_piece()



	def handle_keep_alive(self):
		logger.debug(f'Keep-Alive from {self.Peer}')
		self.artifacts.pop('keep_alive')


	async def handle_choke(self):
		await self.Peer.disconnect(f"Choked client!")
		self.artifacts.pop('choke')


	def handle_unchoke(self):
		self.Peer.choking_me = False
		self.Peer.am_interested = True
		logger.debug(f"Unchoke from {self.Peer}")
		self.artifacts.pop('unchoke')


	async def handle_handshake(self):
		message = self.artifacts['handshake']
		if not message or len(message) < 68:
			# if empty or no response, peer is inactive
			# if response is less than 68, wrong response by peer
			await self.Peer.disconnect("Empty/None/Wrong handshake message! ")

		pstrlen, pstr, res, info_hash, peer_id = unpack('>B19sQ20s20s', message)

		if pstrlen != 19 or pstr != b"BitTorrent protocol":
			await self.Peer.disconnect("Invalid pstrlen or pstr! ")

		handshake_response = {
			"pstrlen": pstrlen,
			"pstr": pstr,
			"reserved": res,
			"info_hash": info_hash,
			"peer_id": peer_id,
		}

		self.Peer.has_handshaked = True
		self.Peer.handshake_response = handshake_response

		logger.debug(f"Handshake from {self.Peer}")
		self.artifacts.pop('handshake')


	def handle_bitfield(self):
		# Create bitfield from bitfield message if available
		# If not available, create empty bitfield_message
		if 'bitfield' in self.artifacts:
			message = self.artifacts['bitfield']
			pieces = BitArray(message)
		else:
			num_pieces = len(self.Peer.torrent_info['piece_hashmap'])
			pieces = BitArray(num_pieces)

		# Merge have requests if available
		if 'have' in self.artifacts:
			for piece_num in self.artifacts['have']:
				pieces[piece_num] = True

		# Finally set Peer pieces value to local pieces value
		self.Peer.pieces = pieces
		try:
			if 'have' in self.artifacts: self.artifacts.pop('have')
			if 'bitfield' in self.artifacts: self.artifacts.pop('bitfield')
		except KeyError:
			...
		finally:
			logger.debug(f"Bitfield from {self.Peer}")


	def handle_piece(self):
		# This is the only method which returns any value
		blocks = list()
		for block_info in self.artifacts['pieces']:
			try:
				index, offset, data = block_info
				block = Block(index, offset, data)
				blocks.append(block)
			except TypeError:
				raise TypeError(f"Handler: Failed To Extract Piece sent by {self.Peer}")
			
		self.artifacts.pop('pieces')
		return blocks



import asyncio
import logging
from random import randint
from typing import Literal
from socket import gaierror
from struct import pack, unpack
from ipaddress import IPv4Address
from http.client import HTTPConnection, HTTPSConnection
from urllib.parse import urlparse, urlencode, ParseResult

from aiotorrent.core.util import chunk
from aiotorrent.core.bencode_utils import bencode_util


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


PEER_ID = b"aiotorrent-XXXXXXXXX"

#TODO: Aggregate global variables in a seperate config file / dictionary

class TrackerBaseClass:
	def __init__(self, tracker_addr: ParseResult, torrent_info: dict) -> None:
		# get first element from the array
		# tracker url is in string format
		tracker_addr = urlparse(tracker_addr)

		self.scheme: str = tracker_addr.scheme
		self.hostname: str = tracker_addr.hostname
		self.port: int = tracker_addr.port
		self.key: int = randint(10000, 99999)

		self.active: bool = False
		# Missing typehint
		self.torrent_info = torrent_info

		self.peers = []
		self.connect_response: dict = {}
		self.announce_response: dict = {}


	def gen_connect_udp(self) -> dict:
		connection_id = 0x41727101980
		action = 0
		transaction_id = randint(1, 65536)

		connect_params = {
			'connection_id': connection_id,
			'action': action,
			'transaction_id': transaction_id
		}

		return connect_params

	@staticmethod
	def serialize_connect(format: Literal["bytes"] | Literal["url"], connect_params: dict) -> bytes | dict:
		if	format == "bytes":
			return pack(
				'>QII',
				connect_params['connection_id'],
				connect_params['action'],
				connect_params['transaction_id'],
			)

		else: return urlencode(connect_params)



	def gen_announce_udp(self, connection_id: int = 0, transaction_id: int = 0) -> dict:
		# Create a dictionary with all the properties of
		# the UDP connect message

		announce_params = {
			'connection_id': connection_id,					# 64 bit integer
			'action': 1,									# 32 bit integer; announce
			'transaction_id': transaction_id,				# 32 bit integer
			'info_hash': self.torrent_info['info_hash'],	# 20 byte string
			'peer_id': PEER_ID,								# 20 byte string; Should be the same and only change if the client restarts
			'downloaded': 0,								# 64 bit integer
			'left': self.torrent_info['size'],				# 64 bit integer
			'uploaded': 0,									# 64 bit integer
			'event': 2,										# 32 bit integer; started
			'ip_address': 0,								# 32 bit integer; 0 is default
			'key': self.key,								# 32 bit integer
			'num_want': 200,								# 32 bit integer; -1 is default
			'port': 6887									# 16 bit integer; should be between 6881 & 6889
		}

		return announce_params

	def gen_announce_http(self):
		# no_peer_id: Not included because this option is ignored if compact is enabled

		# ip: Not included because this option is only needed in the case where the IP address
		# that the request came in on is not the IP address of the client.

		# key, trackerid: Optional, so not included

		announce_params = {
			'info_hash': self.torrent_info['info_hash'],
			'peer_id': PEER_ID,
			'port': 6887,
			'uploaded': 0,
			'downloaded': 0,
			'left': self.torrent_info['size'],
			'compact': 1,
			'event': 'started',
			'num_want': 200
		}

		return announce_params

	def serialize_announce(format: Literal["bytes"] | Literal["url"], announce_params: dict) -> bytes | dict:
		# Pack the message before sending. Packing breakdown
		# Q -> 64 bit integer
		# I -> 32 bit integer
		# s	-> 1 byte string[]
		# 20s -> 20 byte string[]
		# i -> 32 bit unsigned integer
		# H -> 16 bit integer

		if format == "bytes":
			return pack(
			'>QII20s20sQQQIIIiH',
			announce_params['connection_id'],
			announce_params['action'],
			announce_params['transaction_id'],
			announce_params['info_hash'],
			announce_params['peer_id'],
			announce_params['downloaded'],
			announce_params['left'],
			announce_params['uploaded'],
			announce_params['event'],
			announce_params['ip_address'],
			announce_params['key'],
			announce_params['num_want'],
			announce_params['port']
		)
	
		else: return urlencode(announce_params)


	def parse_connect(self, response: bytes) -> dict[str, int]:
		action, transaction_id, connection_id = unpack('>IIQ', response)

		connect_response = {
			'action': action,
			'transaction_id': transaction_id,
			'connection_id': connection_id
		}

		return connect_response
	

	def parse_announce(self, response: bytes) -> dict[str, int | list[tuple[str, int]]]:
		# seperate the properties (20 bytes / 5 properties) and IP list
		response, raw_IPs = response[:20], response[20:]

		# extract the variables with known length
		action, transaction_id, interval, leechers, seeders = unpack('>IIIII', response)

		ip_addresses = list()

		# iterate over the raw_IPs variable going over 6 bytes in each iteration
		# first 4 bytes is IP in decimal format and last 2 bytes is Port in integer format
		for ip_addr in chunk(raw_IPs, 6):
			ip, port = unpack('>IH', ip_addr)
			# convert ip from decimal to dotted format
			ip = IPv4Address(ip).compressed
			ip_addresses.append((ip, port))

		# Add peers received from the tracker to self
		self.peers = ip_addresses

		announce_response = {
			'action': action,
			'transaction_id': transaction_id,
			'interval': interval,
			'leechers': leechers,
			'seeders': seeders,
			'ip_addresses': ip_addresses
		}

		return announce_response





class UDPTracker(TrackerBaseClass):
	# Missing Typehint
	def __init__(self, tracker_addr: str, torrent_info) -> None:
		super().__init__(tracker_addr, torrent_info)


	def __repr__(self) -> str:
		return f"UDPTracker({self.hostname}:{self.port})"


	class UDPProtocolFactory(asyncio.DatagramProtocol):
		# Missing typehint
		def __init__(self, parent_obj):
			self.transport: None | asyncio.DatagramTransport = None
			self.address = (parent_obj.hostname, parent_obj.port)
			self.parent_obj = parent_obj


		def connection_made(self, transport: asyncio.DatagramTransport) -> None:
			self.transport = transport
			# Send connection request when connected
			connect_req_params = self.parent_obj.gen_connect_udp()
			connect_req = UDPTracker.serialize_connect("bytes", connect_req_params)
			logger.debug(f"{self.parent_obj} Sending connect request to {self.address}")
			self.transport.sendto(connect_req)


		def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
			# Check if it's a connect response or an announce response and parse accordingly
			action = unpack('>I', data[:4])
			action = action[0]	# Unpack returns a tuple

			if action == 0:
				# if action is 0, then it's a connect response
				logger.debug(f"{self.parent_obj} Received connect response from {addr}")
				logger.debug(f"{self.parent_obj} Connect response: {data[:16]}")
				self.parent_obj.connect_response = self.parent_obj.parse_connect(data)
				self.parent_obj.active = True

				# prepare announce request and send it
				connection_id = self.parent_obj.connect_response['connection_id']
				transaction_id = self.parent_obj.connect_response['transaction_id']
				announce_req_params = self.parent_obj.gen_announce_udp(connection_id, transaction_id)
				announce_req = UDPTracker.serialize_announce("bytes", announce_req_params)

				# Send announce request
				logger.debug(f"{self.parent_obj} Sending announce request to {self.address}")
				self.transport.sendto(announce_req)


			if action == 1:
				# If action is 1, then it's an announce response
				logger.info(f"{self.parent_obj} Received announce response from {addr}")
				logger.debug(f"{self.parent_obj} Announce response: {data[:32]}")
				self.parent_obj.announce_response = self.parent_obj.parse_announce(data)


	async def get_peers(self, timeout: int = 3) -> list[tuple[str, int]]:

		try:
			loop = asyncio.get_running_loop()
			await loop.create_datagram_endpoint(
				lambda: self.UDPProtocolFactory(self),
				remote_addr = (self.hostname, self.port)
			)
			# This is necessary to wait for the response.
			# We can use asyncio.Event() for this as well
			await asyncio.sleep(timeout)

		except gaierror as e:
			logger.warning(f"Failed to get address info for {self}")
		except Exception as e:
			self.peers = []
			logger.error(e)

		return self.peers



class HTTPTracker(TrackerBaseClass):
	def __init__(self, tracker_addr: str, torrent_info):
		# self.active = False
		super().__init__(tracker_addr, torrent_info)

		# Set http path parameter manually because the BaseTracker doesn't
		self.path = urlparse(tracker_addr).path


	def __repr__(self):
		return f"HTTPTracker({self.hostname}:{self.port})"


	async def get_peers(self):
		self.peers = []
		# It is acceptable with HTTP trackers to directly send the announce request,
		# without sending a connect request first (to reduce network overhead or some other reason)
		announce_req_raw = self.gen_announce_http()
		announce_req = HTTPTracker.serialize_announce("url", announce_req_raw)

		def connect_to_tracker(payload: str) -> dict:
			http_conn_factory = HTTPSConnection if self.scheme == "https" else HTTPConnection
			connection = http_conn_factory(host=self.hostname, port=self.port)
			final_query = f"{self.path}?{payload}"

			#TODO: Add useragent to the http request 
			connection.request("GET", final_query)
			response = connection.getresponse()

			if response.status == 200:
				self.active = True
				self.announce_response = bencode_util.bdecode(response.read())
				peer_list = self.announce_response['peers']
				for ip_addr in chunk(peer_list, 6):
					ip, port = unpack('>IH', ip_addr)
					ip = IPv4Address(ip).compressed
					self.peers.append((ip, port))

			else:
				logger.warning(f"Error fetching peers from {self}: Error Code: {response.status} - {response.reason}: {response.read()}")
				return []

		try:
			loop = asyncio.get_running_loop()
			result = await loop.run_in_executor(None, connect_to_tracker, announce_req)
			return result

		except Exception as e:
			logger.error(f"Error occured while connecting to {self}: {e}")
			return []



class WSSTracker(TrackerBaseClass):
	def __init__(self, *args):
		self.active = False
		self.peers = []
		logger.debug(f"	wss object skipped")


	async def get_peers(self):
		# await asyncio.sleep(1)

		return []




import asyncio
import logging
from enum import Enum
from pathlib import Path


BLOCK_SIZE = 2 ** 14
MIN_BLOCKS_PER_CYCLE = 8
BLOCKS_PER_CYCLE = MIN_BLOCKS_PER_CYCLE

class DownloadStrategy(Enum):
	DEFAULT = 0
	SEQUENTIAL = 10


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Block:
	def __init__(self, piece_num=-1, offset=-1, data=bytes()):
		self.piece_num = piece_num # Piece which this block belongs to
		self.offset = offset
		self.data = data
		self.num = int(offset / BLOCK_SIZE) # Block index of a particular piece


	def __repr__(self):
		return (f"Block #{self.piece_num}-{self.num}")



class PieceWriter:
	def __init__(self, directory_name, file):
		self.file = file
		self.directory = directory_name
		# Create a directory with the same name as torrent name to download files to
		Path(self.directory).mkdir(exist_ok=True)


	def __enter__(self):
		filepath = f"{self.directory}/{self.file.name}"
		self.target_file = open(filepath, 'wb')
		return self


	def write(self, piece):
		# Zero based index of piece within the file
		piece_index = piece.num - self.file.start_piece

		# Calculate offset of piece within the file
		# Reduce file.start_byte from offset in case the file does not start at offset 0 of the piece.
		# In other words, the piece has parts of more than one file.
		offset = (piece_index * piece.piece_size) - self.file.start_byte

		# A negative offset means that this piece contained data for other files
		# Reset offset to 0 in case of a negative offset
		if offset < 0: offset = 0

		# Move the file pointer to the correct offset and write the piece data
		self.target_file.seek(offset)
		self.target_file.write(piece.data)
		logger.info(f"Wrote {piece} to {self.file.name}")


	def __exit__(self, exc_type, exc_value, exc_traceback):
		self.target_file.close()



def chunk(string, size):
	"""A function that splits a string into specified chunk size

		string: str
			The string to be broken / chunked down
		size: int
			The size of each chunk
	"""
	# Use the built-in slice function
	for _ in range(0, len(string), size):
		yield string[:size]
		string = string[size:]



class SequentialPieceDispatcher:
	MEMORY_LIMIT = 1024 * 1024  * 64 # 64MB
	pieces_queue = asyncio.PriorityQueue()
	_pieces_in_queue = 0
	_file_piece_size = 0


	def __init__(self, file, piece_len) -> None:
		self.current = file.start_piece
		self.end = file.end_piece
		self._file_piece_size = piece_len


	async def put(self, piece) -> None:
		queue_size = self.pieces_queue.qsize() * self._file_piece_size
		if queue_size > self.MEMORY_LIMIT:
			raise MemoryError("Memory limit exceeded")
		else:
			await self.pieces_queue.put((piece.num, piece))
			self._pieces_in_queue += 1


	async def dispatch(self):
		while not self.pieces_queue.empty():
			priority, piece = await self.pieces_queue.get()
			if piece.num != self.current:
				self.pieces_queue.put_nowait((priority, piece))
				break
			else:
				self.current += 1
				yield piece


	async def drain(self):
		while not self.pieces_queue.empty():
			_, piece = await self.pieces_queue.get()
			yield piece



from aiotorrent.aiotorrent import Torrent
from aiotorrent.core.util import DownloadStrategy


# Define what is accessible when "from aiotorrent import *" is used
__all__ = ["Torrent", "DownloadStrategy"]



from urllib.parse import urlparse

from aiotorrent.core.trackers import UDPTracker, HTTPTracker, WSSTracker


class TrackerFactory:
	"""
	Using factory method to return the type of tracker required
	"""
	def __new__(self, tracker_addr, torrent_info):

		tracker_types = {
			'udp': UDPTracker,
			'wss': WSSTracker,
			'http': HTTPTracker,
			'https': HTTPTracker,
		}

		t_type = urlparse(tracker_addr).scheme
		return tracker_types[t_type](tracker_addr, torrent_info)